""" Contains the paths to the different models served by onegov.org. """

from datetime import date
from onegov.core.converters import extended_date_converter
from onegov.event import (
    Event,
    EventCollection,
    Occurrence,
    OccurrenceCollection
)
from onegov.form import (
    FormDefinition,
    FormCollection,
    FormSubmissionFile,
    CompleteFormSubmission,
    PendingFormSubmission
)
from onegov.newsletter import (
    Newsletter,
    NewsletterCollection,
    Subscription,
    RecipientCollection,
)
from onegov.org.app import OrgApp
from onegov.org.models import (
    AtoZPages,
    Clipboard,
    Editor,
    Export,
    ExportCollection,
    FormPersonMove,
    GeneralFileCollection,
    ImageFileCollection,
    ImageSet,
    ImageSetCollection,
    LegacyFile,
    LegacyFileCollection,
    LegacyImage,
    LegacyImageCollection,
    News,
    Organisation,
    PageMove,
    PagePersonMove,
    ResourcePersonMove,
    ResourceRecipient,
    ResourceRecipientCollection,
    Search,
    SiteCollection,
    Topic,
)
from onegov.page import PageCollection
from onegov.pay import PaymentProvider, Payment, PaymentCollection
from onegov.pay import PaymentProviderCollection
from onegov.people import Person, PersonCollection
from onegov.reservation import Allocation
from onegov.reservation import Reservation
from onegov.reservation import Resource
from onegov.reservation import ResourceCollection
from onegov.ticket import Ticket, TicketCollection
from onegov.user import Auth, User, UserCollection
from webob import exc
from uuid import UUID


@OrgApp.path(model=Organisation, path='/')
def get_org(app):
    return app.org


@OrgApp.path(model=Auth, path='/auth', converters=dict(skip=bool))
def get_auth(app, to='/', skip=False):
    return Auth.from_app(app, to, skip)


@OrgApp.path(model=User, path='/benutzer/{id}', converters=dict(id=UUID))
def get_user(app, id):
    return UserCollection(app.session()).by_id(id)


@OrgApp.path(model=UserCollection, path='/benutzerverwaltung')
def get_users(app):
    return UserCollection(app.session())


@OrgApp.path(model=Topic, path='/themen', absorb=True)
def get_topic(app, absorb):
    return PageCollection(app.session()).by_path(absorb, ensure_type='topic')


@OrgApp.path(model=News, path='/aktuelles', absorb=True)
def get_news(app, absorb):
    absorb = '/{}/{}'.format('aktuelles', absorb)
    return PageCollection(app.session()).by_path(absorb, ensure_type='news')


@OrgApp.path(model=GeneralFileCollection, path='/dateien')
def get_files(app):
    return GeneralFileCollection(app.session())


@OrgApp.path(model=ImageFileCollection, path='/bilder')
def get_images(app):
    return ImageFileCollection(app.session())


@OrgApp.path(model=ExportCollection, path='/exporte')
def get_exports(request, app):
    return ExportCollection(app)


@OrgApp.path(model=Export, path='/export/{id}')
def get_export(request, app, id):
    return ExportCollection(app).by_id(id)


@OrgApp.path(model=FormCollection, path='/formulare')
def get_forms(app):
    return FormCollection(app.session())


@OrgApp.path(model=FormDefinition, path='/formular/{name}')
def get_form(app, name):
    return FormCollection(app.session()).definitions.by_name(name)


@OrgApp.path(model=PendingFormSubmission, path='/formular-eingabe/{id}',
             converters=dict(id=UUID))
def get_pending_form_submission(app, id):
    return FormCollection(app.session()).submissions.by_id(
        id, state='pending', current_only=True)


@OrgApp.path(model=CompleteFormSubmission, path='/formular-eingang/{id}',
             converters=dict(id=UUID))
def get_complete_form_submission(app, id):
    return FormCollection(app.session()).submissions.by_id(
        id, state='complete', current_only=False)


@OrgApp.path(model=FormSubmissionFile, path='/formular-datei/{id}',
             converters=dict(id=UUID))
def get_form_submission_file(app, id):
    return FormCollection(app.session()).submissions.file_by_id(id)


@OrgApp.path(model=Editor, path='/editor/{action}/{trait}/{page_id}')
def get_editor(app, action, trait, page_id):
    if not Editor.is_supported_action(action):
        return None

    page = PageCollection(app.session()).by_id(page_id)

    if not page.is_supported_trait(trait):
        return None

    if page is not None:
        return Editor(action=action, page=page, trait=trait)


@OrgApp.path(model=PersonCollection, path='/personen')
def get_people(app):
    return PersonCollection(app.session())


@OrgApp.path(model=Person, path='/person/{id}', converters=dict(id=UUID))
def get_person(app, id):
    return PersonCollection(app.session()).by_id(id)


@OrgApp.path(model=Ticket, path='/ticket/{handler_code}/{id}',
             converters=dict(id=UUID))
def get_ticket(app, handler_code, id):
    return TicketCollection(app.session()).by_id(
        id, ensure_handler_code=handler_code)


@OrgApp.path(model=TicketCollection, path='/tickets/{handler}/{state}')
def get_tickets(app, handler='ALL', state='open', page=0, group=None,
                owner=None, extra_parameters=None):
    return TicketCollection(
        app.session(),
        handler=handler,
        state=state,
        page=page,
        group=group,
        owner=owner or '*',
        extra_parameters=extra_parameters
    )


@OrgApp.path(model=ResourceCollection, path='/ressourcen')
def get_resources(app):
    return app.libres_resources


@OrgApp.path(model=Resource, path='/ressource/{name}', converters=dict(
             date=date, highlights_min=int, highlights_max=int))
def get_resource(app, name, date=None, view='agendaWeek',
                 highlights_min=None, highlights_max=None):

    resource = app.libres_resources.by_name(name)

    if resource:
        resource.date = date
        resource.highlights_min = highlights_min
        resource.highlights_max = highlights_max
        resource.view = view

    return resource


@OrgApp.path(model=Allocation, path='/einteilung/{resource}/{id}',
             converters=dict(resource=UUID))
def get_allocation(app, resource, id):
    resource = app.libres_resources.by_id(resource)

    if resource:
        allocation = resource.scheduler.allocations_by_ids((id, )).first()

        # always get the master, even if another id is requested
        return allocation and allocation.get_master()


@OrgApp.path(model=Reservation, path='/reservation/{resource}/{id}',
             converters=dict(resource=UUID))
def get_reservation(app, resource, id):
    resource = app.libres_resources.by_id(resource)

    if resource:
        query = resource.scheduler.managed_reservations()
        query = query.filter(Reservation.id == id)

        return query.first()


@OrgApp.path(model=Clipboard, path='/clipboard/copy/{token}')
def get_clipboard(request, token):
    clipboard = Clipboard(request, token)

    # the url is None if the token is invalid
    if clipboard.url:
        return clipboard


@OrgApp.path(model=SiteCollection, path='/sitecollection')
def get_sitecollection(app):
    return SiteCollection(app.session())


@OrgApp.path(model=PageMove,
             path='/move/page/{subject_id}/{direction}/{target_id}',
             converters=dict(subject_id=int, target_id=int))
def get_page_move(app, subject_id, direction, target_id):
    if subject_id == target_id:
        raise exc.HTTPBadRequest()

    session = app.session()
    pages = PageCollection(session)

    subject = pages.by_id(subject_id)
    target = pages.by_id(target_id)

    if subject and target:
        return PageMove(session, subject, target, direction)


@OrgApp.path(model=PagePersonMove,
             path='/move/page-person/{key}/{subject}/{direction}/{target}')
def get_person_move(app, key, subject, direction, target):
    if subject == target:
        raise exc.HTTPBadRequest()

    session = app.session()
    page = PageCollection(session).by_id(key)

    if page:
        return PagePersonMove(session, page, subject, target, direction)


@OrgApp.path(model=FormPersonMove,
             path='/move/form-person/{key}/{subject}/{direction}/{target}')
def get_form_move(app, key, subject, direction, target):
    session = app.session()
    form = FormCollection(session).definitions.by_name(key)

    if form:
        return FormPersonMove(session, form, subject, target, direction)


@OrgApp.path(
    model=ResourcePersonMove,
    path='/move/resource-person/{key}/{subject}/{direction}/{target}')
def get_resource_move(app, key, subject, direction, target):
    session = app.session()
    resource = ResourceCollection(app.libres_context).by_id(key)

    if resource:
        return ResourcePersonMove(
            session, resource, subject, target, direction)


@OrgApp.path(model=OccurrenceCollection, path='/veranstaltungen',
             converters=dict(start=extended_date_converter,
                             end=extended_date_converter, tags=[]))
def get_occurrences(app, page=0, start=None, end=None, tags=None):
    return OccurrenceCollection(app.session(), page, start, end, tags)


@OrgApp.path(model=Occurrence, path='/veranstaltung/{name}')
def get_occurrence(app, name):
    return OccurrenceCollection(app.session()).by_name(name)


@OrgApp.path(model=EventCollection, path='/events')
def get_events(app, page=0, state='published'):
    return EventCollection(app.session(), page, state)


@OrgApp.path(model=Event, path='/event/{name}')
def get_event(app, name):
    return EventCollection(app.session()).by_name(name)


@OrgApp.path(model=Search, path='/suche')
def get_search(request, q='', page=0):
    return Search(request, q, page)


@OrgApp.path(model=AtoZPages, path='/a-z')
def get_a_to_z(request):
    return AtoZPages(request)


@OrgApp.path(model=NewsletterCollection, path='/newsletters')
def get_newsletters(app):
    return NewsletterCollection(app.session())


@OrgApp.path(model=Newsletter, path='/newsletter/{name}')
def get_newsletter(app, name):
    return get_newsletters(app).by_name(name)


@OrgApp.path(model=RecipientCollection, path='/abonnenten')
def get_newsletter_recipients(app):
    return RecipientCollection(app.session())


@OrgApp.path(model=Subscription, path='/abonnement/{recipient_id}/{token}',
             converters=dict(recipient_id=UUID))
def get_subscription(app, recipient_id, token):
    recipient = RecipientCollection(app.session()).by_id(recipient_id)
    return recipient and Subscription(recipient, token)


@OrgApp.path(model=LegacyFile, path='/datei/{filename}')
def get_file(app, filename):
    return LegacyFileCollection(app).get_file_by_filename(filename)


@OrgApp.path(model=LegacyImage, path='/bild/{filename}')
def get_image(app, filename):
    return LegacyImageCollection(app).get_file_by_filename(filename)


@OrgApp.path(model=ImageSetCollection, path='/fotoalben')
def get_image_sets(app):
    return ImageSetCollection(app.session())


@OrgApp.path(model=ImageSet, path='/fotoalbum/{id}')
def get_image_set(app, id):
    return ImageSetCollection(app.session()).by_id(id)


@OrgApp.path(
    model=ResourceRecipientCollection,
    path='/ressourcen-empfang')
def get_resource_recipient_collection(app):
    return ResourceRecipientCollection(app.session())


@OrgApp.path(
    model=ResourceRecipient,
    path='/ressourcen-empfaenger/{id}',
    converters=dict(id=UUID))
def get_resource_recipient(app, id):
    return ResourceRecipientCollection(app.session()).by_id(id)


@OrgApp.path(
    model=PaymentProviderCollection,
    path='/zahlungsanbieter')
def get_payment_provider_collection(app):
    if app.payment_providers_enabled:
        return PaymentProviderCollection(app.session())


@OrgApp.path(
    model=PaymentProvider,
    path='/zahlungsanbieter-eintrag/{id}',
    converters=dict(id=UUID))
def get_payment_provider(app, id):
    if app.payment_providers_enabled:
        return PaymentProviderCollection(app.session()).by_id(id)


@OrgApp.path(
    model=Payment,
    path='/zahlung/{id}',
    converters=dict(id=UUID))
def get_payment(app, id):
    return PaymentCollection(app.session()).by_id(id)


@OrgApp.path(
    model=PaymentCollection,
    path='/zahlungen')
def get_payments(app, source='*', page=0):
    return PaymentCollection(app.session(), source, page)



from onegov.org import _, OrgApp
from onegov.org.models import GeneralFileCollection, ImageFileCollection
from onegov.org.new_elements import Link, LinkGroup
from onegov.pay import PaymentProviderCollection, PaymentCollection
from onegov.ticket import TicketCollection
from onegov.user.auth import Auth, UserCollection


@OrgApp.template_variables()
def get_template_variables(request):
    return {
        'global_tools': tuple(get_global_tools(request))
    }


def get_global_tools(request):

    # Authentication / Userprofile
    if request.is_logged_in:
        yield LinkGroup(request.current_username, classes=('user', ), links=(
            Link(
                _("User Profile"), request.link(
                    request.app.org, name='benutzerprofil'
                ), attrs={'class': 'profile'}
            ),
            Link(
                _("Logout"), request.link(
                    Auth.from_request_path(request), name='logout'
                ), attrs={'class': 'logout'}
            ),
        ))
    else:
        yield Link(
            _("Login"), request.link(
                Auth.from_request_path(request), name='login'
            ), attrs={'class': 'login'}
        )

        if request.app.settings.org.enable_user_registration:
            yield Link(
                _("Register"), request.link(
                    Auth.from_request_path(request), name='register'
                ), attrs={'class': 'register'})

    # Management dropdown
    if request.is_manager:
        links = []

        links.append(
            Link(
                _("Files"), request.class_link(GeneralFileCollection),
                attrs={'class': 'files'}
            )
        )

        links.append(
            Link(
                _("Images"), request.class_link(ImageFileCollection),
                attrs={'class': 'images'}
            )
        )

        if request.is_admin and request.app.payment_providers_enabled:
            links.append(
                Link(
                    _("Payment Provider"),
                    request.class_link(PaymentProviderCollection),
                    attrs={'class': 'payment-provider'}
                )
            )

        links.append(
            Link(
                _("Payments"),
                request.class_link(PaymentCollection),
                attrs={'class': 'payment'}
            )
        )

        if request.is_admin:
            links.append(
                Link(
                    _("Settings"), request.link(
                        request.app.org, 'einstellungen'
                    ), attrs={'class': 'settings'}
                )
            )

            links.append(
                Link(
                    _("Users"), request.class_link(UserCollection),
                    attrs={'class': 'users'}
                )
            )

        yield LinkGroup(_("Management"), classes=('management', ), links=links)

    # Tickets
    if request.is_manager:
        ticket_count = request.app.ticket_count
        screen_count = ticket_count.open or ticket_count.pending

        links = []

        links.append(
            Link(
                _("Open Tickets"),
                request.class_link(
                    TicketCollection, {'handler': 'ALL', 'state': 'open'}
                ),
                attrs={
                    'class': ('with-count', 'alert', 'open-tickets'),
                    'data-count': str(ticket_count.open)
                }
            )
        )

        links.append(
            Link(
                _("Pending Tickets"),
                request.class_link(
                    TicketCollection, {'handler': 'ALL', 'state': 'pending'}
                ),
                attrs={
                    'class': ('with-count', 'info', 'pending-tickets'),
                    'data-count': str(ticket_count.pending)
                }
            )
        )

        links.append(
            Link(
                _("Closed Tickets"),
                url=request.class_link(
                    TicketCollection, {'handler': 'ALL', 'state': 'closed'}
                ),
                attrs={
                    'class': ('with-count', 'secondary', 'closed-tickets'),
                    'data-count': str(ticket_count.closed),
                }
            )
        )

        if screen_count:
            css = ticket_count.open and 'alert' or 'info'
        else:
            css = 'no-tickets'

        yield LinkGroup(
            screen_count == 1 and _("Ticket") or _("Tickets"),
            classes=('with-count', css),
            links=links,
            attributes={'data-count': str(screen_count)}
        )



""" Contains the models describing files and images. """

from collections import namedtuple
from datetime import datetime
from dateutil.relativedelta import relativedelta
from itertools import groupby
from onegov.core.orm.mixins import meta_property
from onegov.file import File, FileSet, FileCollection, FileSetCollection
from onegov.file.utils import IMAGE_MIME_TYPES_AND_SVG
from onegov.org import _
from onegov.org.models.extensions import HiddenFromPublicExtension
from onegov.search import ORMSearchable
from sedate import standardize_date, utcnow
from sqlalchemy import desc


DateInterval = namedtuple('DateInterval', ('name', 'start', 'end'))


class GroupFilesByDateMixin(object):

    def get_date_intervals(self, today):
        today = standardize_date(today, 'UTC')

        month_end = today + relativedelta(day=31)
        month_start = today - relativedelta(day=1)

        yield DateInterval(
            name=_("This month"),
            start=month_start,
            end=month_end)

        last_month_end = month_start - relativedelta(microseconds=1)
        last_month_start = month_start - relativedelta(months=1)

        yield DateInterval(
            name=_("Last month"),
            start=last_month_start,
            end=last_month_end)

        if month_start.month not in (1, 2):
            this_year_end = last_month_start - relativedelta(microseconds=1)
            this_year_start = this_year_end.replace(
                month=1, day=1, hour=0, minute=0, second=0, microsecond=0)

            yield DateInterval(
                name=_("This year"),
                start=this_year_start,
                end=this_year_end)
        else:
            this_year_end = None
            this_year_start = None

        last_year_end = this_year_start or last_month_start
        last_year_end -= relativedelta(microseconds=1)
        last_year_start = last_year_end.replace(
            month=1, day=1, hour=0, minute=0, second=0, microsecond=0)

        yield DateInterval(
            name=_("Last year"),
            start=last_year_start,
            end=last_year_end)

        older_end = last_year_start - relativedelta(microseconds=1)
        older_start = datetime(2000, 1, 1, tzinfo=today.tzinfo)

        yield DateInterval(
            name=_("Older"),
            start=older_start,
            end=older_end)

    def query_intervals(self, intervals, before_filter=None, process=None):
        base_query = self.query().order_by(desc(File.last_change))

        if before_filter:
            base_query = before_filter(base_query)

        for interval in intervals:
            query = base_query.filter(File.last_change >= interval.start)
            query = query.filter(File.last_change <= interval.end)

            for result in query.all():
                if process is not None:
                    yield interval.name, process(result)

    def grouped_by_date(self, today=None, id_only=True):
        """ Returns all files grouped by natural language dates.

        By default, only ids are returned, as this is enough to build the
        necessary links, which is what you usually want from a file.

        The given date is expected to be in UTC.

        """

        intervals = tuple(self.get_date_intervals(today or utcnow()))

        def before_filter(query):
            if id_only:
                return query.with_entities(File.id)

            return query

        def process(result):
            if id_only:
                return result.id

            return result

        return groupby(
            self.query_intervals(intervals, before_filter, process),
            key=lambda item: item[0]
        )


class GeneralFile(File):
    __mapper_args__ = {'polymorphic_identity': 'general'}


class ImageFile(File):
    __mapper_args__ = {'polymorphic_identity': 'image'}


class ImageSet(FileSet, HiddenFromPublicExtension, ORMSearchable):
    __mapper_args__ = {'polymorphic_identity': 'image'}

    es_properties = {
        'title': {'type': 'localized'},
        'lead': {'type': 'localized'}
    }

    @property
    def es_public(self):
        return not self.is_hidden_from_public

    @property
    def es_language(self):
        return 'de'  # xxx for now there's no other language

    @property
    def es_suggestions(self):
        return {
            "input": [self.title.lower()]
        }

    lead = meta_property('lead')
    view = meta_property('view')

    show_images_on_homepage = meta_property('show_images_on_homepage')


class ImageSetCollection(FileSetCollection):

    def __init__(self, session):
        super().__init__(session, type='image')


class GeneralFileCollection(FileCollection, GroupFilesByDateMixin):

    supported_content_types = 'all'

    def __init__(self, session):
        super().__init__(session, type='general', allow_duplicates=False)


class ImageFileCollection(FileCollection, GroupFilesByDateMixin):

    supported_content_types = IMAGE_MIME_TYPES_AND_SVG

    def __init__(self, session):
        super().__init__(session, type='image', allow_duplicates=False)



from onegov.form import Form, merge_forms
from onegov.form.validators import ValidFormDefinition
from onegov.org import _
from onegov.org.forms.fields import HtmlField
from onegov.org.forms.generic import DateRangeForm
from onegov.org.forms.generic import ExportForm
from onegov.org.forms.generic import PaymentMethodForm
from onegov.org.forms.reservation import RESERVED_FIELDS
from wtforms import RadioField
from wtforms import StringField
from wtforms import TextAreaField
from wtforms import TextField
from wtforms import validators
from wtforms.fields.html5 import DecimalField


def as_float(value):
    if not value:
        return 0.0

    return float(value)


class ResourceBaseForm(Form):
    """ Defines the form for all resources. """
    title = StringField(_("Title"), [validators.InputRequired()])

    lead = TextAreaField(
        label=_("Lead"),
        description=_("Describes what this reservation resource is about"),
        render_kw={'rows': 4})

    group = StringField(
        label=_("Group"),
        description=_("Used to group the resource in the overview")
    )

    text = HtmlField(
        label=_("Text"))

    definition = TextAreaField(
        label=_("Extra Fields Definition"),
        validators=[
            validators.Optional(),
            ValidFormDefinition(
                require_email_field=False,
                reserved_fields=RESERVED_FIELDS
            )
        ],
        render_kw={'rows': 32, 'data-editor': 'form'}
    )

    pricing_method = RadioField(
        label=_("Price"),
        fieldset=_("Payments"),
        default='free',
        validators=[validators.InputRequired()],
        choices=(
            ('free', _("Free of charge")),
            ('per_item', _("Per item")),
            ('per_hour', _("Per hour"))
        )
    )

    price_per_item = DecimalField(
        label=_("Price per item"),
        filters=(as_float, ),
        fieldset=_("Payments"),
        validators=[validators.Optional()],
        depends_on=('pricing_method', 'per_item')
    )

    price_per_hour = DecimalField(
        label=_("Price per hour"),
        filters=(as_float, ),
        fieldset=_("Payments"),
        validators=[validators.Optional()],
        depends_on=('pricing_method', 'per_hour')
    )

    currency = TextField(
        label=_("Currency"),
        default="CHF",
        fieldset=_("Payments"),
        depends_on=('pricing_method', '!free'),
        validators=[validators.InputRequired()],
    )

    def ensure_valid_price(self):
        if self.pricing_method.data == 'per_item':
            if not float(self.price_per_item.data) > 0:
                self.price_per_item.errors.append(_(
                    "The price must be larger than zero"
                ))
                return False

        if self.pricing_method.data == 'per_hour':
            if not float(self.price_per_hour.data) > 0:
                self.price_per_hour.errors.append(_(
                    "The price must be larger than zero"
                ))
                return False


class ResourceForm(merge_forms(ResourceBaseForm, PaymentMethodForm)):
    pass


class ResourceCleanupForm(DateRangeForm):
    """ Defines the form to remove multiple allocations. """


class ResourceExportForm(merge_forms(DateRangeForm, ExportForm)):
    """ Resource export form with start/end date. """



from onegov.core import utils
from onegov.core.orm.mixins import content_property
from onegov.org import _
from onegov.org.forms import LinkForm, PageForm
from onegov.org.models.atoz import AtoZ
from onegov.org.models.extensions import ContactExtension
from onegov.org.models.extensions import CoordinatesExtension
from onegov.org.models.extensions import HiddenFromPublicExtension
from onegov.org.models.extensions import PersonLinkExtension
from onegov.org.models.extensions import VisibleOnHomepageExtension
from onegov.org.models.traitinfo import TraitInfo
from onegov.page import Page
from onegov.search import ORMSearchable
from sqlalchemy import desc, func
from sqlalchemy.dialects.postgresql import JSON
from sqlalchemy.orm import undefer, object_session


class SearchablePage(ORMSearchable):

    es_properties = {
        'title': {'type': 'localized'},
        'lead': {'type': 'localized'},
        'text': {'type': 'localized_html'}
    }

    lead = content_property('lead')
    text = content_property('text')
    url = content_property('url')

    @property
    def es_public(self):
        return not self.is_hidden_from_public

    @property
    def es_language(self):
        return 'de'  # xxx for now there's no other language

    @property
    def es_skip(self):
        return self.meta.get('trait') == 'link'  # do not index links

    @property
    def es_suggestions(self):
        return {
            "input": [self.title.lower()]
        }


class Topic(Page, TraitInfo, SearchablePage, HiddenFromPublicExtension,
            VisibleOnHomepageExtension, ContactExtension, PersonLinkExtension,
            CoordinatesExtension):
    __mapper_args__ = {'polymorphic_identity': 'topic'}

    es_type_name = 'topics'

    @property
    def deletable(self):
        """ Returns true if this page may be deleted. """
        return self.parent is not None

    @property
    def editable(self):
        return True

    @property
    def paste_target(self):
        if self.trait == 'link':
            return self.parent

        if self.trait == 'page':
            return self

        raise NotImplementedError

    @property
    def allowed_subtraits(self):
        if self.trait == 'link':
            return tuple()

        if self.trait == 'page':
            return ('page', 'link')

        raise NotImplementedError

    def is_supported_trait(self, trait):
        return trait in {'link', 'page'}

    def get_form_class(self, trait, request):
        if trait == 'link':
            return self.with_content_extensions(LinkForm, request, extensions=[
                HiddenFromPublicExtension,
                VisibleOnHomepageExtension
            ])

        if trait == 'page':
            return self.with_content_extensions(PageForm, request)

        raise NotImplementedError


class News(Page, TraitInfo, SearchablePage, HiddenFromPublicExtension,
           VisibleOnHomepageExtension, ContactExtension, PersonLinkExtension,
           CoordinatesExtension):
    __mapper_args__ = {'polymorphic_identity': 'news'}

    es_type_name = 'news'

    @property
    def absorb(self):
        return utils.lchop(self.path, 'aktuelles').lstrip('/')

    @property
    def deletable(self):
        return self.parent is not None

    @property
    def editable(self):
        return self.parent is not None

    @property
    def paste_target(self):
        if self.parent:
            return self.parent
        else:
            return self

    @property
    def allowed_subtraits(self):
        # only allow one level of news
        if self.parent is None:
            return ('news', )
        else:
            return tuple()

    def is_supported_trait(self, trait):
        return trait in {'news'}

    def get_form_class(self, trait, request):
        if trait == 'news':
            form_class = self.with_content_extensions(PageForm, request)

            if hasattr(form_class, 'is_visible_on_homepage'):
                # clarify the intent of this particular flag on the news, as
                # the effect is not entirely the same (news may be shown on the
                # homepage anyway)
                form_class.is_visible_on_homepage.kwargs['label'] = _(
                    "Always visible on homepage")

            return form_class

        raise NotImplementedError

    def news_query(self, limit=2):
        news = object_session(self).query(News)
        news = news.filter(Page.parent == self)
        news = news.order_by(desc(Page.created))
        news = news.options(undefer('created'))
        news = news.options(undefer('content'))
        news = news.limit(limit)

        sticky = func.json_extract_path_text(
            func.cast(Page.meta, JSON), 'is_visible_on_homepage') == 'true'

        sticky_news = news.limit(None)
        sticky_news = sticky_news.filter(sticky)

        return news.union(sticky_news).order_by(desc(Page.created))

    @property
    def years(self):
        query = object_session(self).query(News)
        query = query.with_entities(func.date_part('year', Page.created))
        query = query.group_by(func.date_part('year', Page.created))
        query = query.filter(Page.parent == self)

        return sorted([int(r[0]) for r in query.all()], reverse=True)


class AtoZPages(AtoZ):

    def get_title(self, item):
        return item.title

    def get_items(self):

        # XXX implement correct collation support on the database level
        topics = self.request.app.session().query(Topic).all()
        topics = sorted(topics, key=self.sortkey)

        if self.request.is_manager:
            return [topic for topic in topics if topic.trait == 'page']
        else:
            return [
                topic for topic in topics if topic.trait == 'page' and
                not topic.is_hidden_from_public
            ]



from onegov.page import Page
from onegov.form import FormDefinition
from onegov.reservation import Resource


class PersonMove(object):
    """ Represents a single move of a linked person. """

    def __init__(self, session, obj, subject, target, direction):
        self.session = session
        self.obj = obj

        # remove the dashes from the uuid
        self.subject = subject.replace('-', '')
        self.target = target.replace('-', '')
        self.direction = direction

    def execute(self):
        self.obj.move_person(
            subject=self.subject,
            target=self.target,
            direction=self.direction
        )

    @classmethod
    def for_url_template(cls, obj):
        return cls(
            session=None,
            obj=obj,
            subject='{subject_id}',
            target='{target_id}',
            direction='{direction}'
        )

    @staticmethod
    def get_implementation(obj):
        # XXX this is not really extendable by other org applications. They
        # need to override this function *and* define new paths for each class
        # they add.
        #
        # To simplify we should create a registry using dectate, which handles
        # the classes for us. We'd only have one path and the PersonMove
        # class would gain a 'type' and a generic 'key' property which we would
        # use to lookup the actual handler for the move.
        #
        # For example:
        #
        # @App.person_move_handler()
        # class PagePersonMoveMapping(object):
        #     def get_object(self, key):
        #        ...
        #     def get_key(self, obj):
        #        ...
        #
        if isinstance(obj, Page):
            return PagePersonMove

        if isinstance(obj, FormDefinition):
            return FormPersonMove

        if isinstance(obj, Resource):
            return ResourcePersonMove

        raise NotImplementedError


class PagePersonMove(PersonMove):
    """ Represents a single move of a linked person on a page. """

    @property
    def key(self):
        return self.obj.id


class FormPersonMove(PersonMove):
    """ Represents a single move of a linked person on a form definition. """

    @property
    def key(self):
        return self.obj.name


class ResourcePersonMove(PersonMove):
    """ Represents a single move of a linked person on a form definition. """

    @property
    def key(self):
        return self.obj.id



from onegov.core.security import Private
from onegov.org import OrgApp, _
from onegov.org.layout import PaymentCollectionLayout
from onegov.pay import Payment
from onegov.pay import PaymentCollection
from onegov.pay import PaymentProviderCollection
from onegov.ticket import TicketCollection


@OrgApp.html(
    model=PaymentCollection,
    template='payments.pt',
    permission=Private)
def view_payments(self, request):
    session = request.app.session()
    tickets = TicketCollection(session)

    def ticket(link):
        if link.__tablename__ == 'reservations':
            return tickets.by_handler_id(link.token.hex)
        elif link.__tablename__ == 'submissions':
            return tickets.by_handler_id(link.id.hex)
        else:
            raise NotImplementedError

    providers = {
        provider.id: provider
        for provider in PaymentProviderCollection(session).query()
    }

    payment_links = self.payment_links_by_batch()

    return {
        'title': _("Payments"),
        'layout': PaymentCollectionLayout(self, request),
        'payments': self.batch,
        'get_ticket': ticket,
        'providers': providers,
        'payment_links': payment_links
    }


@OrgApp.view(
    model=Payment,
    name='mark-as-paid',
    request_method='POST',
    permission=Private)
def mark_as_paid(self, request):
    request.assert_valid_csrf_token()

    assert self.source == 'manual'
    self.state = 'paid'


@OrgApp.view(
    model=Payment,
    name='mark-as-unpaid',
    request_method='POST',
    permission=Private)
def mark_as_unpaid(self, request):
    request.assert_valid_csrf_token()

    assert self.source == 'manual'
    self.state = 'open'


@OrgApp.view(
    model=Payment,
    name='capture',
    request_method='POST',
    permission=Private)
def capture(self, request):
    request.assert_valid_csrf_token()

    assert self.source == 'stripe_connect'
    self.charge.capture()


@OrgApp.view(
    model=Payment,
    name='refund',
    request_method='POST',
    permission=Private)
def refund(self, request):
    request.assert_valid_csrf_token()

    assert self.source == 'stripe_connect'
    self.refund()



from collections import OrderedDict

from onegov.foundation import BaseTheme
from onegov.core.utils import module_path


# options editable by the user
user_options = {
    'primary-color': '#006fba',
    'footer-height': '38px'
}


class OrgTheme(BaseTheme):
    name = 'onegov.org.foundation'

    # don't touch this number, it's incremented using bumpversion, so every
    # release will automatically trigger a rebuild of the theme
    version = '0.10.0'

    @property
    def default_options(self):
        options = OrderedDict((
            # tile images
            ('tile-image-1', '"../static/homepage-images/tile-1-small.jpg"'),
            ('tile-image-2', '"../static/homepage-images/tile-2-small.jpg"'),
            ('tile-image-3', '"../static/homepage-images/tile-3-small.jpg"'),
            ('tile-image-4', '"../static/homepage-images/tile-4-small.jpg"'),
            ('tile-image-5', '"../static/homepage-images/tile-5-small.jpg"'),
            ('tile-image-6', '"../static/homepage-images/tile-6-small.jpg"'),
        ))
        options.update(user_options)

        return options

    @property
    def foundation_components(self):
        return (
            'grid',
            'alert-boxes',
            'block-grid',
            'breadcrumbs',
            'button-groups',
            'buttons',
            'dropdown',
            'dropdown-buttons',
            'forms',
            'inline-lists',
            'labels',
            'orbit',
            'pagination',
            'panels',
            'progress-bars',
            'reveal',
            'side-nav',
            'split-buttons',
            'sub-nav',
            'tables',
            'thumbs',
            'tooltips',
            'top-bar',
            'type',
            'visibility',
        )

    @property
    def pre_imports(self):
        return [
            'foundation-mods',
        ]

    @property
    def post_imports(self):
        return [
            'org'
        ]

    @property
    def extra_search_paths(self):
        return [module_path('onegov.org.theme', 'styles')]



from onegov.core.html import sanitize_html
from onegov.org.utils import annotate_html
from wtforms.fields import TextAreaField


class HtmlField(TextAreaField):
    """ A textfield with html with integrated sanitation and annotation,
    cleaning the html and adding extra information including setting the
    image size by querying the database.

    """

    def __init__(self, *args, **kwargs):
        self.form = kwargs.get('_form')

        if 'render_kw' not in kwargs or not kwargs['render_kw'].get('class_'):
            kwargs['render_kw'] = kwargs.get('render_kw', {})
            kwargs['render_kw']['class_'] = 'editor'

        super().__init__(*args, **kwargs)

    def pre_validate(self, form):
        self.data = annotate_html(sanitize_html(self.data), form.request)
